# Copyright (c) 2021 Aiven, Helsinki, Finland. https://aiven.io/
from aiven.client.argx import UserError
from enum import Enum

import os


class ConnectionInfoError(UserError):
    def __init__(self, message):
        super().__init__()
        self.message = message

    def __str__(self):
        return self.message


class Store(Enum):
    overwrite = object()
    write = object()
    skip = object()

    def handle(self, getter, path):
        if self is Store.overwrite:
            write = True
        elif self is Store.write:
            write = not os.path.exists(path)
        elif self is Store.skip:
            write = False
        else:
            raise NotImplementedError

        if write:
            value = getter()
            with open(path, "w", encoding="utf-8") as fob:
                fob.write(value)



# Copyright 2015, Aiven, https://aiven.io/
#
# This file is under the Apache License, Version 2.0.
# See the file `LICENSE` for details.

# pylint: disable=no-member
from aiven.client.session import AivenClientAdapter, get_requests_session
from requests import Session

import pytest


def test_valid_requests_session():
    """Test that get_requests_session returns a valid Session that has the expected parameters set.
    """

    session = get_requests_session()

    assert isinstance(session, Session)
    assert "aiven-client" in session.headers["User-Agent"]

    adapter = session.adapters["https://"]
    assert isinstance(adapter, AivenClientAdapter)
    assert hasattr(adapter, "timeout")
    assert adapter.timeout is None


@pytest.mark.parametrize("argument,value", [
    ("timeout", 30),
    ("timeout", 0),
])
def test_adapter_parameters_are_passed_along(argument, value):
    session = get_requests_session(**{argument: value})
    adapter = session.adapters["https://"]
    assert isinstance(adapter, AivenClientAdapter)
    assert hasattr(adapter, argument)
    assert getattr(adapter, argument) == value



# Copyright (c) 2021 Aiven, Helsinki, Finland. https://aiven.io/
from ._utils import find_component, find_user, format_uri
from .common import ConnectionInfoError


class PGConnectionInfo:
    def __init__(self, host, port, username, dbname, password, sslmode):
        self.host = host
        self.port = port
        self.username = username
        self.dbname = dbname
        self.password = password
        self.sslmode = sslmode

    @classmethod
    def from_service(cls, service, *, route, usage, privatelink_connection_id, username, dbname, sslmode):
        if service["service_type"] != "pg":
            raise ConnectionInfoError("Cannot format pg connection info for service type {service_type}".format_map(service))

        info = find_component(
            service["components"], route=route, usage=usage, privatelink_connection_id=privatelink_connection_id
        )
        host = info["host"]
        port = info["port"]
        user = find_user(service, username)
        password = user.get("password")
        if password is None:
            raise ConnectionInfoError(f"Could not find password for username {username}")
        return cls(host=host, port=port, username=username, dbname=dbname, password=password, sslmode=sslmode)

    def params(self):
        return {
            "host": self.host,
            "port": self.port,
            "user": self.username,
            "dbname": self.dbname,
            "password": self.password,
            "sslmode": self.sslmode,
        }

    def uri(self):
        return format_uri(
            scheme="postgres",
            username=self.username,
            password=self.password,
            host=self.host,
            port=self.port,
            path=f"/{self.dbname}",
            query={"sslmode": self.sslmode},
        )

    def connection_string(self):
        return f"host='{self.host}' port='{self.port}' user={self.username} dbname='{self.dbname}'"

    def psql(self):
        return ["psql", self.uri()]



from requests import adapters, Session
from requests.structures import CaseInsensitiveDict
from typing import Optional

try:
    from .version import __version__  # pylint: disable=no-name-in-module
except ImportError:
    __version__ = "UNKNOWN"


class AivenClientAdapter(adapters.HTTPAdapter):
    def __init__(self, *args, timeout: Optional[int] = None, **kwargs):
        self.timeout = timeout
        super().__init__(*args, **kwargs)

    def send(self, *args, **kwargs):  # pylint: disable=signature-differs
        if not kwargs.get("timeout"):
            kwargs["timeout"] = self.timeout
        return super().send(*args, **kwargs)


def get_requests_session(*, timeout: Optional[int] = None) -> Session:
    adapter = AivenClientAdapter(timeout=timeout)

    session = Session()
    session.mount("http://", adapter)
    session.mount("https://", adapter)
    session.verify = True
    session.headers = CaseInsensitiveDict({
        "content-type": "application/json",
        "user-agent": "aiven-client/" + __version__,
    })

    return session



# Copyright 2021, Aiven, https://aiven.io/
#
# This file is under the Apache License, Version 2.0.
# See the file `LICENSE` for details.
from aiven.client.speller import suggest

import pytest

# Current service types (2021-03-09); they are used just for testing
SERVICE_TYPES = [
    "alerta",
    "cassandra",
    "elasticsearch",
    "grafana",
    "influxdb",
    "kafka",
    "kafka_connect",
    "kafka_mirrormaker",
    "m3db",
    "m3coordinator",
    "m3aggregator",
    "mysql",
    "pg",
    "redis",
    "sw",
    "flink",
]


@pytest.mark.parametrize(
    ["word_to_check", "known_words", "suggestion"],
    [
        ("kafka", SERVICE_TYPES, "kafka"),
        ("kakfa", SERVICE_TYPES, "kafka"),
        ("kafkaconnect", SERVICE_TYPES, "kafka_connect"),
        ("kafka-connect", SERVICE_TYPES, "kafka_connect"),
        ("asdf", SERVICE_TYPES, None),
    ],
)
def test_suggest(word_to_check, known_words, suggestion):
    assert suggest(word_to_check=word_to_check, known_words=known_words) == suggestion



"""
automatically maintains the latest git tag + revision info in a python file

"""

import importlib.machinery
import os
import subprocess


def get_project_version(version_file):
    version_file = os.path.join(os.path.dirname(os.path.realpath(__file__)), version_file)
    try:
        module = importlib.machinery.SourceFileLoader("version", version_file).load_module()
        file_ver = module.__version__
    except:  # pylint: disable=bare-except
        file_ver = None

    try:
        proc = subprocess.Popen(
            ["git", "describe", "--always"],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        stdout, _ = proc.communicate()
        if stdout:
            git_ver = stdout.splitlines()[0].strip().decode("utf-8")
            if git_ver and ((git_ver != file_ver) or not file_ver):
                open(version_file, "w").write("__version__ = '%s'\n" % git_ver)
                return git_ver
    except OSError:
        pass

    if not file_ver:
        raise Exception("version not available from git or from file %r" % version_file)

    return file_ver


if __name__ == "__main__":
    import sys

    get_project_version(sys.argv[1])



# Spelling Corrector in Python 3; see http://norvig.com/spell-correct.html
#
# Copyright (c) 2007-2016 Peter Norvig
# MIT license: www.opensource.org/licenses/mit-license.php


def suggest(word_to_check, known_words):
    def get_correction(word):
        """Most probable spelling correction for word."""
        candidates = get_candidates(word)
        return next(iter(candidates)) if candidates else None

    def get_candidates(word):
        """Generate possible spelling corrections for word."""
        return get_known([word]) or get_known(get_edits1(word)) or get_known(get_edits2(word)) or []

    def get_known(words):
        """The subset of `words` that appear in the dictionary of WORDS."""
        return set(w for w in words if w in known_words)

    def get_edits1(word):
        """All edits that are one edit away from `word`."""
        letters = "abcdefghijklmnopqrstuvwxyz_"
        splits = [(word[:i], word[i:]) for i in range(len(word) + 1)]
        deletes = [L + R[1:] for L, R in splits if R]
        transposes = [L + R[1] + R[0] + R[2:] for L, R in splits if len(R) > 1]
        replaces = [L + c + R[1:] for L, R in splits if R for c in letters]
        inserts = [L + c + R for L, R in splits for c in letters]
        return set(deletes + transposes + replaces + inserts)

    def get_edits2(word):
        """All edits that are two edits away from `word`."""
        return (e2 for e1 in get_edits1(word) for e2 in get_edits1(e1))

    return get_correction(word_to_check)



# Copyright 2020, Aiven, https://aiven.io/
#
# This file is under the Apache License, Version 2.0.
# See the file `LICENSE` for details.
try:
    from functools import cached_property
except ImportError:
    cached_property = None

from aiven.client.argx import arg, CommandLineTool


class TestCLI(CommandLineTool):
    @arg()
    def xxx(self):
        """7"""

    @arg()
    def aaa(self):
        """1"""

    @arg()
    def ccc(self):
        """4"""


class SubCLI(CommandLineTool):
    @arg()
    def yyy(self):
        """8"""

    @arg()
    def bbb(self):
        """2"""

    @arg()
    def ddd(self):
        """5"""


class SubCLI2(CommandLineTool):
    @arg()
    def yyz(self):
        """9"""

    @arg()
    def bbc(self):
        """3"""

    @arg()
    def dde(self):
        """6"""


def test_extended_commands_remain_alphabetically_ordered():
    cli = TestCLI("testcli")
    cli.extend_commands(cli)  # Force the CLI to have its full arg set at execution

    sl2 = SubCLI2("subcli2")
    sl = SubCLI("subcli")

    cli.extend_commands(sl2)
    cli.extend_commands(sl)

    action_order = [item.dest for item in cli.subparsers._choices_actions]  # pylint: disable=protected-access
    assert action_order == ["aaa", "bbb", "bbc", "ccc", "ddd", "dde", "xxx", "yyy", "yyz"]


class DescriptorCLI(CommandLineTool):
    @property
    def raise1(self):
        raise RuntimeError("evaluated raise1")

    if cached_property is not None:

        @cached_property
        def raise2(self):
            raise RuntimeError("evaluated raise2")

    @arg("something")
    def example_command(self):
        """Example command."""


def test_descriptors_are_not_eagerly_evaluated():
    cli = DescriptorCLI("DescriptorCLI")
    calls = []
    cli.add_cmds(calls.append)
    assert calls == [cli.example_command]



# Copyright 2015, Aiven, https://aiven.io/
# coding=utf-8
#
# This file is under the Apache License, Version 2.0.
# See the file `LICENSE` for details.
"""Pretty-print JSON objects and lists as tables"""
import datetime
import decimal
import fnmatch
import ipaddress
import json
import sys

# string type checking must work on python 2.x and 3.x
try:
    basestring
except NameError:
    basestring = str  # pylint: disable=redefined-builtin


class CustomJsonEncoder(json.JSONEncoder):
    def default(self, o):  # pylint:disable=E0202
        if isinstance(o, (datetime.datetime, datetime.date)):
            return o.isoformat()
        if isinstance(o, datetime.timedelta):
            return str(o)
        if isinstance(o, decimal.Decimal):
            return str(o)
        if isinstance(
            o, (
                ipaddress.IPv4Address, ipaddress.IPv6Address, ipaddress.IPv4Network, ipaddress.IPv6Network,
                ipaddress.IPv4Interface, ipaddress.IPv6Interface
            )
        ):
            return o.compressed

        return json.JSONEncoder.default(self, o)


def format_item(key, value):
    if isinstance(value, list):
        formatted = ", ".join(format_item(None, entry) for entry in value)
    elif isinstance(value, dict):
        formatted = json.dumps(value, sort_keys=True, cls=CustomJsonEncoder)
    elif isinstance(value, basestring):
        if key and key.endswith("_time") and value.endswith("Z") and "." in value:
            # drop microseconds from timestamps
            value = value.split(".", 1)[0] + "Z"
        # json encode strings, but if the input string is exactly the same
        # as the output without quotes we'll go with the original
        json_v = json.dumps(value)
        quoted_v = '"{}"'.format(value)
        if (json_v == quoted_v or json_v.replace("\\u00a3", "£").replace("\\u20ac", "€") == quoted_v):
            formatted = value
        else:
            formatted = json_v
    elif isinstance(value, datetime.datetime):
        formatted = value.isoformat()
    elif isinstance(value, datetime.timedelta):
        formatted = str(value)
    else:
        # again, if adding quotes is only thing json econding would do, omit them
        json_v = json.dumps(value, sort_keys=True, cls=CustomJsonEncoder)
        quoted_v = '"{}"'.format(value)
        if json_v == quoted_v:
            formatted = "{}".format(value)
        else:
            formatted = json_v

    return formatted


def yield_table(result, drop_fields=None, table_layout=None, header=True):
    """format a list of dicts in a nicer table format yielding string rows"""
    if not result:
        return

    if not isinstance(result[0], dict):
        for item in result:
            yield format_item(None, item)
        return

    drop_fields = set(drop_fields or [])

    def iter_values(key, value):
        if not isinstance(value, dict):
            yield key, value
            return
        for subkey, subvalue in value.items():
            for kv in iter_values((key + "." if key else "") + subkey, subvalue):
                yield kv

    # format all fields and collect their widths
    widths = {}
    formatted_values = []
    for item in result:
        formatted_row = {}
        formatted_values.append(formatted_row)
        for key, value in item.items():
            if key not in drop_fields:
                for subkey, subvalue in iter_values(key, value):
                    formatted_row[subkey] = format_item(subkey, subvalue)
                    widths[subkey] = max(len(subkey), len(formatted_row[subkey]), widths.get(subkey, 1))

    # default table layout is one row per item with sorted field names
    if table_layout is None:
        table_layout = sorted(widths)
    if not isinstance(table_layout[0], (list, tuple)):
        table_layout = [table_layout]

    horizontal_fields = table_layout[0]
    if header:
        yield "  ".join(f.upper().ljust(widths[f]) for f in horizontal_fields)
        yield "  ".join("=" * widths[f] for f in horizontal_fields)
    for row_num, formatted_row in enumerate(formatted_values):
        # If we have multiple lines per entry yield an empty line between each entry
        if len(table_layout) > 1 and row_num > 0:
            yield ""
        # The main, horizontal, line
        yield "  ".join(formatted_row.get(f, "").ljust(widths[f]) for f in horizontal_fields).strip()
        # And the rest of the fields, one per field
        fields_to_print = []
        for vertical_field in table_layout[1:]:
            if vertical_field.endswith(".*"):
                for key, value in sorted(formatted_row.items()):
                    if fnmatch.fnmatch(key, vertical_field):
                        fields_to_print.append((key, value))
            else:
                value = formatted_row.get(vertical_field)
                if value is not None:
                    fields_to_print.append((vertical_field, value))
        if fields_to_print:
            max_key_width = max(len(key) for key, _ in fields_to_print)
            for key, value in fields_to_print:
                yield "    {:{}} = {}".format(key, max_key_width, value)


def print_table(result, drop_fields=None, table_layout=None, header=True, file=None):  # pylint: disable=redefined-builtin
    """print a list of dicts in a nicer table format"""
    for row in yield_table(result, drop_fields=drop_fields, table_layout=table_layout, header=header):
        print(row, file=file or sys.stdout)



# Copyright 2015, Aiven, https://aiven.io/
#
# This file is under the Apache License, Version 2.0.
# See the file `LICENSE` for details.
"""
Configurable parameters via environment variables
"""

import os

USER_HOME = os.path.expanduser("~")

AIVEN_CONFIG_DIR = os.environ.get("AIVEN_CONFIG_DIR", os.path.join(USER_HOME, ".config", "aiven"))

AIVEN_AUTH_TOKEN = os.environ.get("AIVEN_AUTH_TOKEN")
AIVEN_CA_CERT = os.environ.get("AIVEN_CA_CERT")
AIVEN_CLIENT_CONFIG = os.environ.get("AIVEN_CLIENT_CONFIG", os.path.join(AIVEN_CONFIG_DIR, "aiven-client.json"))
AIVEN_CREDENTIALS_FILE = os.environ.get("AIVEN_CREDENTIALS_FILE", os.path.join(AIVEN_CONFIG_DIR, "aiven-credentials.json"))
AIVEN_PROJECT = os.environ.get("AIVEN_PROJECT")
AIVEN_WEB_URL = os.environ.get("AIVEN_WEB_URL")



# Copyright 2015, Aiven, https://aiven.io/
#
# This file is under the Apache License, Version 2.0.
# See the file `LICENSE` for details.
from .common import UNDEFINED
from .session import get_requests_session
from urllib.parse import quote

import json
import logging
import re
import requests
import time
import warnings

try:
    from .version import __version__  # pylint: disable=no-name-in-module
except ImportError:
    __version__ = "UNKNOWN"


class Error(Exception):
    """Request error"""

    def __init__(self, response, status=520):
        Exception.__init__(self, response.text)
        self.response = response
        self.status = status


class AivenClientBase:  # pylint: disable=old-style-class
    """Aiven Client with low-level HTTP operations"""

    def __init__(self, base_url, show_http=False, request_timeout=None):
        self.log = logging.getLogger("AivenClient")
        self.auth_token = None
        self.base_url = base_url
        self.log.debug("using %r", self.base_url)
        self.session = get_requests_session(timeout=request_timeout)
        self.http_log = logging.getLogger("aiven_http")
        self.init_http_logging(show_http)
        self.api_prefix = "/v1"

    def init_http_logging(self, show_http):
        http_handler = logging.StreamHandler()
        http_handler.setFormatter(logging.Formatter("%(message)s"))
        self.http_log.addHandler(http_handler)
        self.http_log.propagate = False
        if show_http:
            self.http_log.setLevel(logging.DEBUG)

    def set_auth_token(self, token):
        self.auth_token = token

    def set_ca(self, ca):
        self.session.verify = ca

    def _execute(self, func, method, path, body, params=None):
        url = self.base_url + path
        headers = {}
        if isinstance(body, dict):
            headers["content-type"] = "application/json"
            data = json.dumps(body)
            log_data = json.dumps(body, sort_keys=True, indent=4)
        else:
            headers["content-type"] = "application/octet-stream"
            data = body
            log_data = data or ""

        if self.auth_token:
            headers["authorization"] = "aivenv1 {token}".format(token=self.auth_token)

        self.http_log.debug("-----Request Begin-----")
        self.http_log.debug("%s %s %s", method, url, params if params else "")
        for header, header_value in headers.items():
            self.http_log.debug("%s: %s", header, header_value)

        self.http_log.debug("")
        self.http_log.debug("%s", log_data)
        self.http_log.debug("-----Request End-----")

        response = func(url, headers=headers, params=params, data=data)

        self.http_log.debug("-----Response Begin-----")
        self.http_log.debug("%s %s", response.status_code, response.reason)
        for header, header_value in response.headers.items():
            self.http_log.debug("%s: %s", header, header_value)

        self.http_log.debug("")
        if response.headers.get("content-type") == "application/json":
            self.http_log.debug("%s", json.dumps(response.json(), sort_keys=True, indent=4))
        else:
            self.http_log.debug("%s", response.text)

        self.http_log.debug("-----Response End-----")

        if not str(response.status_code).startswith("2"):
            raise Error(response, status=response.status_code)

        return response

    def get(self, path="", params=None):
        """HTTP GET"""
        return self._execute(self.session.get, "GET", path, body=None, params=params)

    def patch(self, path="", body=None, params=None):
        """HTTP PATCH"""
        return self._execute(self.session.patch, "PATCH", path, body, params)

    def post(self, path="", body=None, params=None):
        """HTTP POST"""
        return self._execute(self.session.post, "POST", path, body, params)

    def put(self, path="", body=None, params=None):
        """HTTP PUT"""
        return self._execute(self.session.put, "PUT", path, body, params)

    def delete(self, path="", body=None, params=None):
        """HTTP DELETE"""
        return self._execute(self.session.delete, "DELETE", path, body, params)

    def verify(self, op, path, body=None, params=None, result_key=None, retry=None):
        # Retry GET operations by default
        if retry is None and op == self.get:  # pylint: disable=comparison-with-callable
            attempts = 3
        else:
            attempts = 1 + (retry or 0)

        path = self.api_prefix + path

        while attempts:
            attempts -= 1
            try:
                if body is not None:
                    response = op(path=path, body=body, params=params)
                else:
                    response = op(path=path, params=params)
                break
            except requests.exceptions.ConnectionError as ex:
                if attempts <= 0:
                    raise
                self.log.warning(
                    "%s %s failed: %s: %s; retrying in 0.2 seconds, %s attempts left",
                    op.__name__.upper(),
                    path,
                    ex.__class__.__name__,
                    ex,
                    attempts,
                )
                time.sleep(0.2)

        result = response.json()
        if result.get("error"):
            raise Error(
                "server returned error: {op} {base_url}{path} {result}".format(
                    op=op.__doc__, base_url=self.base_url, path=path, result=result
                )
            )

        if result_key is not None:
            return result[result_key]
        return result

    @staticmethod
    def build_path(*parts):
        return "/" + "/".join(quote(part, safe="") for part in parts)


class AivenClient(AivenClientBase):
    """Aiven Client with high-level operations"""

    def get_service_versions(self):
        return self.verify(self.get, "/service_versions", result_key="service_versions")

    def get_service_indexes(self, project, service):
        return self.verify(
            self.get,
            self.build_path("project", project, "service", service, "index"),
            result_key="indexes",
        )

    def delete_service_index(self, project, service, index_name):
        return self.verify(
            self.delete,
            self.build_path("project", project, "service", service, "index", index_name),
        )

    def get_clouds(self, project):
        if project is None:
            path = "/clouds"
        else:
            path = self.build_path("project", project, "clouds")
        return self.verify(self.get, path, result_key="clouds")

    def get_service(self, project, service):
        return self.verify(
            self.get,
            self.build_path("project", project, "service", service),
            result_key="service",
        )

    def get_service_metrics(self, project, service, period):
        path = self.build_path("project", project, "service", service, "metrics")
        return self.verify(self.post, path, result_key="metrics", body={"period": period})

    def authenticate_user(self, email, password, otp=None, tenant_id=None):
        body = {
            "email": email,
            "password": password,
        }
        if otp is not None:
            body["otp"] = otp
        if tenant_id is not None:
            body["tenant"] = tenant_id

        return self.verify(self.post, "/userauth", body=body)

    def create_service_connection_pool(
        self,
        project,
        service,
        pool_name,
        dbname,
        username,
        pool_size=None,
        pool_mode=None,
    ):
        body = {"database": dbname, "username": username, "pool_name": pool_name}
        if pool_size:
            body["pool_size"] = pool_size
        if pool_mode:
            body["pool_mode"] = pool_mode
        return self.verify(
            self.post,
            self.build_path("project", project, "service", service, "connection_pool"),
            body=body,
        )

    def update_service_connection_pool(
        self,
        project,
        service,
        pool_name,
        dbname=None,
        username=None,
        pool_size=None,
        pool_mode=None,
    ):
        body = {}
        if username is not None:
            body["username"] = username
        if dbname is not None:
            body["database"] = dbname
        if pool_size is not None:
            body["pool_size"] = pool_size
        if pool_mode is not None:
            body["pool_mode"] = pool_mode
        path = self.build_path("project", project, "service", service, "connection_pool", pool_name)
        return self.verify(self.put, path, body=body)

    def delete_service_connection_pool(self, project, service, pool_name):
        path = self.build_path("project", project, "service", service, "connection_pool", pool_name)
        return self.verify(self.delete, path)

    def create_service_database(self, project, service, dbname):
        return self.verify(
            self.post,
            self.build_path("project", project, "service", service, "db"),
            body={"database": dbname},
        )

    def delete_service_database(self, project, service, dbname):
        path = self.build_path("project", project, "service", service, "db", dbname)
        return self.verify(self.delete, path)

    def create_service_user(self, project, service, username, extra_params=None):
        body = {"username": username}
        if extra_params:
            body.update(extra_params)
        return self.verify(
            self.post,
            self.build_path("project", project, "service", service, "user"),
            body=body,
            result_key="user",
        )

    def delete_service_user(self, project, service, username):
        path = self.build_path("project", project, "service", service, "user", username)
        return self.verify(self.delete, path)

    def get_service_user(self, project, service, username):
        path = self.build_path("project", project, "service", service, "user", username)
        return self.verify(self.get, path, result_key="user")

    def reset_service_user_password(self, project, service, username, password):
        path = self.build_path("project", project, "service", service, "user", username)
        body = {"operation": "reset-credentials"}
        if password is not None:
            body["new_password"] = password
        return self.verify(self.put, path, body=body)

    def set_service_user_access_control(self, project, service, username, access_control):
        path = self.build_path("project", project, "service", service, "user", username)
        body = {"operation": "set-access-control", "access_control": access_control}
        return self.verify(self.put, path, body=body)

    def get_service_integration_endpoints(self, project):
        path = self.build_path("project", project, "integration_endpoint")
        return self.verify(self.get, path, result_key="service_integration_endpoints")

    def get_service_integration_endpoint_types(self, project):
        path = self.build_path("project", project, "integration_endpoint_types")
        return self.verify(self.get, path, result_key="endpoint_types")

    def create_service_integration_endpoint(self, project, endpoint_name, endpoint_type, user_config):
        return self.verify(
            self.post,
            self.build_path("project", project, "integration_endpoint"),
            body={
                "endpoint_name": endpoint_name,
                "endpoint_type": endpoint_type,
                "user_config": user_config,
            },
        )

    def update_service_integration_endpoint(self, project, endpoint_id, user_config):
        return self.verify(
            self.put,
            self.build_path("project", project, "integration_endpoint", endpoint_id),
            body={
                "user_config": user_config,
            },
        )

    def delete_service_integration_endpoint(self, project, endpoint_id):
        return self.verify(
            self.delete,
            self.build_path("project", project, "integration_endpoint", endpoint_id),
        )

    def get_service_integrations(self, project, service):
        path = self.build_path("project", project, "service", service, "integration")
        return self.verify(self.get, path, result_key="service_integrations")

    def get_service_integration_types(self, project):
        path = self.build_path("project", project, "integration_types")
        return self.verify(self.get, path, result_key="integration_types")

    def create_service_integration(
        self,
        project,
        integration_type,
        source_service=None,
        dest_service=None,
        source_endpoint_id=None,
        dest_endpoint_id=None,
        user_config=None,
    ):
        user_config = user_config or {}
        return self.verify(
            self.post,
            self.build_path("project", project, "integration"),
            body={
                "source_endpoint_id": source_endpoint_id,
                "source_service": source_service,
                "dest_endpoint_id": dest_endpoint_id,
                "dest_service": dest_service,
                "integration_type": integration_type,
                "user_config": user_config,
            },
        )

    def update_service_integration(self, project, integration_id, user_config):
        return self.verify(
            self.put,
            self.build_path("project", project, "integration", integration_id),
            body={
                "user_config": user_config,
            },
            result_key="service_integration",
        )

    def get_service_integration(self, project, integration_id):
        path = self.build_path("project", project, "integration", integration_id)
        return self.verify(self.get, path, result_key="service_integration")

    def delete_service_integration(self, project, integration_id):
        return self.verify(
            self.delete,
            self.build_path("project", project, "integration", integration_id),
        )

    def create_service_task(self, project, service, body):
        return self.verify(
            self.post,
            self.build_path("project", project, "service", service, "task"),
            body=body,
        )

    def get_service_task(self, project, service, task_id):
        path = self.build_path("project", project, "service", service, "task", task_id)
        return self.verify(self.get, path, result_key="task")

    def get_service_topic(self, project, service, topic):
        path = self.build_path("project", project, "service", service, "topic", topic)
        return self.verify(self.get, path, result_key="topic")

    def _set_namespace_options(
        self,
        ns,
        ns_ret=None,
        ns_res=None,
        ns_blocksize_dur=None,
        ns_block_data_expiry_dur=None,
        ns_buffer_future_dur=None,
        ns_buffer_past_dur=None,
        ns_writes_to_commitlog=None
    ):
        re_ns_duration = re.compile(r'\d+[smhd]')

        def _validate_ns_dur(val):
            if not re_ns_duration.match(val):
                raise ValueError(f"Invalid namespace duration value '{val}'")

        ns["options"] = ns.get("options", {})
        ns["options"]["retention_options"] = ns["options"].get("retention_options", {})
        if ns_ret:
            _validate_ns_dur(ns_ret)
            ns["options"]["retention_options"]["retention_period_duration"] = ns_ret
        if ns_res:
            _validate_ns_dur(ns_res)
            ns["resolution"] = ns_res
        if ns_blocksize_dur:
            _validate_ns_dur(ns_blocksize_dur)
            ns["options"]["retention_options"]["blocksize_duration"] = ns_blocksize_dur
        if ns_block_data_expiry_dur:
            _validate_ns_dur(ns_block_data_expiry_dur)
            ns["options"]["retention_options"]["block_data_expiry_duration"] = ns_block_data_expiry_dur
        if ns_buffer_future_dur:
            _validate_ns_dur(ns_buffer_future_dur)
            ns["options"]["retention_options"]["buffer_future_duration"] = ns_buffer_future_dur
        if ns_buffer_past_dur:
            _validate_ns_dur(ns_buffer_past_dur)
            ns["options"]["retention_options"]["buffer_past_duration"] = ns_buffer_past_dur
        if ns_writes_to_commitlog is not None:
            ns["options"]["writes_to_commitlog"] = bool(ns_writes_to_commitlog)

    def list_m3_namespaces(self, project, service):
        service_res = self.get_service(project=project, service=service)
        return service_res.get("user_config", {}).get("namespaces", [])

    def delete_m3_namespace(self, project, service, ns_name):
        service_res = self.get_service(project=project, service=service)
        old_namespaces = service_res.get("user_config", {}).get("namespaces", [])
        new_namespaces = [ns for ns in old_namespaces if ns["name"] != ns_name]
        if len(old_namespaces) == len(new_namespaces):
            raise KeyError(f"Namespace '{ns_name}' does not exist")
        return self.verify(
            self.put,
            self.build_path("project", project, "service", service),
            body={"user_config": {
                "namespaces": new_namespaces
            }},
        )

    def add_m3_namespace(
        self,
        project,
        service,
        ns_name,
        ns_type,
        ns_ret,
        ns_res=None,
        ns_blocksize_dur=None,
        ns_block_data_expiry_dur=None,
        ns_buffer_future_dur=None,
        ns_buffer_past_dur=None,
        ns_writes_to_commitlog=None
    ):
        service_res = self.get_service(project=project, service=service)
        namespaces = service_res.get("user_config", {}).get("namespaces", [])
        valid_namespace_types = {'unaggregated', 'aggregated'}
        if ns_type not in valid_namespace_types:
            raise ValueError(f"Invalid namespace type {ns_type}, valid types {valid_namespace_types}")
        new_namespace = {
            "name": ns_name,
            "type": ns_type,
        }
        self._set_namespace_options(
            ns=new_namespace,
            ns_ret=ns_ret,
            ns_res=ns_res,
            ns_blocksize_dur=ns_blocksize_dur,
            ns_block_data_expiry_dur=ns_block_data_expiry_dur,
            ns_buffer_future_dur=ns_buffer_future_dur,
            ns_buffer_past_dur=ns_buffer_past_dur,
            ns_writes_to_commitlog=ns_writes_to_commitlog
        )
        namespaces.append(new_namespace)
        return self.verify(
            self.put,
            self.build_path("project", project, "service", service),
            body={"user_config": {
                "namespaces": namespaces
            }},
        )

    def update_m3_namespace(
        self,
        project,
        service,
        ns_name,
        ns_ret=None,
        ns_res=None,
        ns_blocksize_dur=None,
        ns_block_data_expiry_dur=None,
        ns_buffer_future_dur=None,
        ns_buffer_past_dur=None,
        ns_writes_to_commitlog=None
    ):
        service_res = self.get_service(project=project, service=service)
        namespaces = service_res.get("user_config", {}).get("namespaces", [])
        namespace = None
        for ns in namespaces:
            if ns["name"] == ns_name:
                namespace = ns
        if not namespace:
            raise KeyError(f"Namespace '{ns_name}' does not exist")
        self._set_namespace_options(
            ns=namespace,
            ns_ret=ns_ret,
            ns_res=ns_res,
            ns_blocksize_dur=ns_blocksize_dur,
            ns_block_data_expiry_dur=ns_block_data_expiry_dur,
            ns_buffer_future_dur=ns_buffer_future_dur,
            ns_buffer_past_dur=ns_buffer_past_dur,
            ns_writes_to_commitlog=ns_writes_to_commitlog
        )
        return self.verify(
            self.put,
            self.build_path("project", project, "service", service),
            body={"user_config": {
                "namespaces": namespaces
            }},
        )

    def list_service_topics(self, project, service):
        return self.verify(
            self.get,
            self.build_path("project", project, "service", service, "topic"),
            result_key="topics",
        )

    def create_service_topic(
        self,
        project,
        service,
        topic,
        partitions,
        replication,
        min_insync_replicas,
        retention_bytes,
        retention_hours,
        cleanup_policy,
        tags=None,
    ):
        body = {
            "cleanup_policy": cleanup_policy,
            "min_insync_replicas": min_insync_replicas,
            "topic_name": topic,
            "partitions": partitions,
            "replication": replication,
            "retention_bytes": retention_bytes,
            "retention_hours": retention_hours,
        }
        if tags is not None:
            body.update({"tags": tags})
        return self.verify(
            self.post,
            self.build_path("project", project, "service", service, "topic"),
            body=body,
        )

    def update_service_topic(
        self,
        project,
        service,
        topic,
        partitions,
        retention_bytes,
        retention_hours,
        min_insync_replicas,
        replication=None,
        tags=None,
    ):
        body = {
            "partitions": partitions,
            "min_insync_replicas": min_insync_replicas,
            "replication": replication,
            "retention_bytes": retention_bytes,
            "retention_hours": retention_hours,
        }
        if tags is not None:
            body.update({"tags": tags})
        return self.verify(
            self.put,
            self.build_path("project", project, "service", service, "topic", topic),
            body=body,
        )

    def delete_service_topic(self, project, service, topic):
        return self.verify(
            self.delete,
            self.build_path("project", project, "service", service, "topic", topic),
        )

    def list_service_elasticsearch_acl_config(self, project, service):
        return self.verify(
            self.get,
            self.build_path("project", project, "service", service, "elasticsearch", "acl"),
        )

    @staticmethod
    def _add_es_acl_rules(config, user, rules):
        user_acl = None
        for acl in config["acls"]:
            if acl["username"] == user:
                user_acl = acl
                break
        if user_acl is None:  # new user
            user_acl = {"username": user, "rules": []}
            config["acls"].append(user_acl)
        patterns = {rule["index"]: rule["permission"] for rule in user_acl["rules"]}
        patterns.update(rules)
        user_acl["rules"] = [{"index": index, "permission": permission} for index, permission in patterns.items()]

    @staticmethod
    def _del_es_acl_rules(config, user, rules):
        acls = []
        user_acl = None
        for acl in config["acls"]:
            if acl["username"] != user:
                acls.append(acl)
            if acl["username"] == user:
                user_acl = acl
        config["acls"] = acls
        if user_acl is None:  # No rules existed for the user
            return
        if not rules:  # removing all rules
            return

        # Remove the requested rules
        user_acl["rules"] = [rule for rule in user_acl["rules"] if rule["index"] not in rules]
        config["acls"].append(user_acl)

    def update_service_elasticsearch_acl_config(
        self,
        project,
        service,
        enabled=None,
        extended_acl=None,
        username=None,
        add_rules=None,
        del_rules=None,
    ):
        acl_config = self.list_service_elasticsearch_acl_config(project, service)["elasticsearch_acl_config"]
        if enabled is not None:
            acl_config["enabled"] = enabled
        if extended_acl is not None:
            acl_config["extendedAcl"] = extended_acl
        if add_rules is not None:
            try:
                rules = {index.strip(): permission.strip() for index, permission in [rule.split("/") for rule in add_rules]}
            except ValueError as ex:
                raise ValueError("Unrecognized index-pattern/permission rule") from ex
            self._add_es_acl_rules(config=acl_config, user=username, rules=rules)
        if del_rules is not None:
            self._del_es_acl_rules(
                config=acl_config,
                user=username,
                rules=set(rule.strip() for rule in del_rules),
            )

        path = self.build_path("project", project, "service", service, "elasticsearch", "acl")
        return self.verify(self.put, path, body={"elasticsearch_acl_config": acl_config})

    def add_service_kafka_acl(self, project, service, permission, topic, username):
        return self.verify(
            self.post,
            self.build_path("project", project, "service", service, "acl"),
            body={
                "permission": permission,
                "topic": topic,
                "username": username,
            },
        )

    def create_connector_config_based_on_current(self, project, service, connector_name, config_update):
        current_connectors = self.list_kafka_connectors(project, service)
        connector = [conn for conn in current_connectors["connectors"] if conn["name"] == connector_name]
        if not connector:
            raise KeyError("Current configuration for connector '{}' not in connector list".format(connector_name))
        assert len(connector) == 1
        full_config = connector[0]["config"]
        full_config.update(config_update)
        return full_config

    def delete_service_kafka_acl(self, project, service, acl_id):
        return self.verify(
            self.delete,
            self.build_path("project", project, "service", service, "acl", acl_id),
        )

    def get_available_kafka_connectors(self, project, service):
        return self.verify(
            self.get,
            self.build_path("project", project, "service", service, "available-connectors"),
        )

    def list_kafka_connectors(self, project, service):
        return self.verify(
            self.get,
            self.build_path("project", project, "service", service, "connectors"),
        )

    def get_kafka_connector_status(self, project, service, connector_name):
        path = self.build_path(
            "project",
            project,
            "service",
            service,
            "connectors",
            connector_name,
            "status",
        )
        return self.verify(self.get, path)

    def get_kafka_connector_schema(self, project, service, connector_name):
        path = self.build_path(
            "project",
            project,
            "service",
            service,
            "connector-plugins",
            connector_name,
            "configuration",
        )
        return self.verify(self.get, path)

    def create_kafka_connector(self, project, service, config):
        return self.verify(
            self.post,
            self.build_path("project", project, "service", service, "connectors"),
            body=config,
        )

    def update_kafka_connector(self, project, service, connector_name, config, fetch_current=False):
        path = self.build_path("project", project, "service", service, "connectors", connector_name)
        if fetch_current:
            config = self.create_connector_config_based_on_current(project, service, connector_name, config)
        return self.verify(self.put, path, body=config)

    def delete_kafka_connector(self, project, service, connector_name):
        path = self.build_path("project", project, "service", service, "connectors", connector_name)
        return self.verify(self.delete, path)

    def pause_kafka_connector(self, project, service, connector_name):
        path = self.build_path(
            "project",
            project,
            "service",
            service,
            "connectors",
            connector_name,
            "pause",
        )
        return self.verify(self.post, path)

    def resume_kafka_connector(self, project, service, connector_name):
        path = self.build_path(
            "project",
            project,
            "service",
            service,
            "connectors",
            connector_name,
            "resume",
        )
        return self.verify(self.post, path)

    def restart_kafka_connector(self, project, service, connector_name):
        path = self.build_path(
            "project",
            project,
            "service",
            service,
            "connectors",
            connector_name,
            "restart",
        )
        return self.verify(self.post, path)

    def restart_kafka_connector_task(self, project, service, connector_name, task_id):
        path = self.build_path(
            "project",
            project,
            "service",
            service,
            "connectors",
            connector_name,
            "tasks",
            task_id,
            "restart",
        )
        return self.verify(self.post, path)

    def get_schema(self, project, service, schema_id):
        path = self.build_path(
            "project",
            project,
            "service",
            service,
            "kafka",
            "schema",
            "schemas",
            "ids",
            schema_id,
        )
        return self.verify(self.get, path)

    def check_schema_compatibility(self, project, service, subject, version, schema):
        path = self.build_path(
            "project",
            project,
            "service",
            service,
            "kafka",
            "schema",
            "compatibility",
            "subjects",
            subject,
            "versions",
            version,
        )
        return self.verify(self.post, path, body={"schema": schema})

    def get_schema_global_configuration(self, project, service):
        path = self.build_path("project", project, "service", service, "kafka", "schema", "config")
        return self.verify(self.get, path)

    def update_schema_global_configuration(self, project, service, compatibility):
        path = self.build_path("project", project, "service", service, "kafka", "schema", "config")
        return self.verify(self.put, path, body={"compatibility": compatibility})

    def get_schema_subject_configuration(self, project, service, subject):
        path = self.build_path("project", project, "service", service, "kafka", "schema", "config", subject)
        return self.verify(self.get, path)

    def update_schema_subject_configuration(self, project, service, subject, compatibility):
        path = self.build_path("project", project, "service", service, "kafka", "schema", "config", subject)
        return self.verify(self.put, path, body={"compatibility": compatibility})

    def list_schema_subjects(self, project, service):
        path = self.build_path("project", project, "service", service, "kafka", "schema", "subjects")
        return self.verify(self.get, path)

    def delete_schema_subject(self, project, service, subject):
        path = self.build_path(
            "project",
            project,
            "service",
            service,
            "kafka",
            "schema",
            "subjects",
            subject,
        )
        return self.verify(self.delete, path)

    def get_schema_subject_version(self, project, service, subject, version):
        path = self.build_path(
            "project",
            project,
            "service",
            service,
            "kafka",
            "schema",
            "subjects",
            subject,
            "versions",
            version,
        )
        return self.verify(self.get, path)

    def get_schema_subject_version_schema(self, project, service, subject, version):
        path = self.build_path(
            "project",
            project,
            "service",
            service,
            "kafka",
            "schema",
            "subjects",
            subject,
            "versions",
            version,
            "schema",
        )
        return self.verify(self.get, path)

    def list_schema_subject_versions(self, project, service, subject):
        path = self.build_path(
            "project",
            project,
            "service",
            service,
            "kafka",
            "schema",
            "subjects",
            subject,
            "versions",
        )
        return self.verify(self.get, path)

    def create_schema_subject_version(self, project, service, subject, schema):
        path = self.build_path(
            "project",
            project,
            "service",
            service,
            "kafka",
            "schema",
            "subjects",
            subject,
            "versions",
        )
        return self.verify(self.post, path, body={"schema": schema})

    def delete_schema_subject_version(self, project, service, subject, version):
        path = self.build_path(
            "project",
            project,
            "service",
            service,
            "kafka",
            "schema",
            "subjects",
            subject,
            "versions",
            version,
        )
        return self.verify(self.delete, path)

    def list_mirrormaker_replication_flows(self, project, service):
        path = self.build_path("project", project, "service", service, "mirrormaker", "replication-flows")
        return self.verify(self.get, path, result_key="replication_flows")

    def create_mirrormaker_replication_flow(self, project, service, source_cluster, target_cluster, config):
        path = self.build_path("project", project, "service", service, "mirrormaker", "replication-flows")
        body = {}
        body.update(config)
        body["source_cluster"] = source_cluster
        body["target_cluster"] = target_cluster
        return self.verify(self.post, path, body=body)

    def update_mirrormaker_replication_flow(self, project, service, source_cluster, target_cluster, config):
        path = self.build_path(
            "project",
            project,
            "service",
            service,
            "mirrormaker",
            "replication-flows",
            source_cluster,
            target_cluster,
        )
        body = {}
        body.update(config)
        return self.verify(self.put, path, body=body, result_key="replication_flow")

    def get_mirrormaker_replication_flow(self, project, service, source_cluster, target_cluster):
        path = self.build_path(
            "project",
            project,
            "service",
            service,
            "mirrormaker",
            "replication-flows",
            source_cluster,
            target_cluster,
        )
        return self.verify(self.get, path, result_key="replication_flow")

    def delete_mirrormaker_replication_flow(self, project, service, source_cluster, target_cluster):
        path = self.build_path(
            "project",
            project,
            "service",
            service,
            "mirrormaker",
            "replication-flows",
            source_cluster,
            target_cluster,
        )
        return self.verify(self.delete, path)

    def list_flink_tables(self, project, service):
        path = self.build_path(
            "project",
            project,
            "service",
            service,
            "flink",
            "table",
        )
        return self.verify(
            self.get,
            path,
            result_key="tables",
        )

    def create_flink_table(
        self,
        project,
        service,
        integration_id,
        table_name,
        schema_sql,
        kafka_topic=None,
        jdbc_table=None,
        partitioned_by=None,
        like_options=None
    ):
        path = self.build_path(
            "project",
            project,
            "service",
            service,
            "flink",
            "table",
        )
        body = {
            "integration_id": integration_id,
            "name": table_name,
            "schema_sql": schema_sql,
        }
        if kafka_topic:
            body["kafka_topic"] = kafka_topic
        if jdbc_table:
            body["jdbc_table"] = jdbc_table
        if partitioned_by:
            body["partitioned_by"] = partitioned_by
        if like_options:
            body["like_options"] = like_options
        return self.verify(self.post, path, body=body)

    def get_flink_table(self, project, service, table_id):
        path = self.build_path("project", project, "service", service, "flink", "table", table_id)
        return self.verify(
            self.get,
            path,
        )

    def delete_flink_table(self, project, service, table_id):
        path = self.build_path("project", project, "service", service, "flink", "table", table_id)
        return self.verify(self.delete, path)

    def create_flink_job(
        self,
        project,
        service,
        statement,
        job_name,
        tables=None,
    ):
        if tables is None:
            tables = []
        path = self.build_path(
            "project",
            project,
            "service",
            service,
            "flink",
            "job",
        )
        body = {
            "statement": statement,
            "tables": tables,
            "job_name": job_name,
        }
        return self.verify(self.post, path, body=body)

    def list_flink_jobs(self, project, service):
        path = self.build_path(
            "project",
            project,
            "service",
            service,
            "flink",
            "job",
        )

        return self.verify(
            self.get,
            path,
            result_key="jobs",
        )

    def get_flink_job(self, project, service, job_id):
        path = self.build_path("project", project, "service", service, "flink", "job", job_id)
        return self.verify(
            self.get,
            path,
        )

    def cancel_flink_job(self, project, service, job_id):
        path = self.build_path("project", project, "service", service, "flink", "job", job_id, "cancel")
        return self.verify(
            self.post,
            path,
        )

    def list_project_vpcs(self, project):
        return self.verify(self.get, self.build_path("project", project, "vpcs"))

    def create_project_vpc(self, project, cloud, network_cidr, peering_connections):
        return self.verify(
            self.post,
            self.build_path("project", project, "vpcs"),
            body={
                "cloud_name": cloud,
                "network_cidr": network_cidr,
                "peering_connections": peering_connections,
            },
        )

    def request_project_vpc(self, project, cloud, network_cidr, peering_connections):
        warnings.warn("Use the create_project_vpc method", DeprecationWarning)
        return self.create_project_vpc(
            project=project,
            cloud=cloud,
            network_cidr=network_cidr,
            peering_connections=peering_connections,
        )

    def get_project_vpc(self, project, project_vpc_id):
        return self.verify(self.get, self.build_path("project", project, "vpcs", project_vpc_id))

    def delete_project_vpc(self, project, project_vpc_id):
        return self.verify(self.delete, self.build_path("project", project, "vpcs", project_vpc_id))

    def create_project_vpc_peering_connection(
        self,
        project,
        project_vpc_id,
        peer_cloud_account,
        peer_vpc,
        peer_region=None,
        peer_resource_group=None,
        peer_azure_app_id=None,
        peer_azure_tenant_id=None,
        user_peer_network_cidrs=None,
    ):
        path = self.build_path("project", project, "vpcs", project_vpc_id, "peering-connections")
        body = {
            "peer_cloud_account": peer_cloud_account,
            "peer_vpc": peer_vpc,
        }
        if peer_region is not None:
            body["peer_region"] = peer_region
        if peer_resource_group is not None:
            body["peer_resource_group"] = peer_resource_group
        if peer_azure_app_id is not None:
            body["peer_azure_app_id"] = peer_azure_app_id
        if peer_azure_tenant_id is not None:
            body["peer_azure_tenant_id"] = peer_azure_tenant_id
        if user_peer_network_cidrs is not None:
            body["user_peer_network_cidrs"] = user_peer_network_cidrs
        return self.verify(self.post, path, body=body)

    def request_project_vpc_peering_connection(self, project, project_vpc_id, peer_cloud_account, peer_vpc):
        warnings.warn("Use the create_project_vpc_peering_connection method", DeprecationWarning)
        return self.create_project_vpc_peering_connection(
            project=project,
            project_vpc_id=project_vpc_id,
            peer_cloud_account=peer_cloud_account,
            peer_vpc=peer_vpc,
        )

    def delete_project_vpc_peering_connection(
        self,
        project,
        project_vpc_id,
        peer_cloud_account,
        peer_vpc,
        peer_region=None,
        peer_resource_group=UNDEFINED,
    ):
        path = self.build_path(
            "project",
            project,
            "vpcs",
            project_vpc_id,
            "peering-connections",
            "peer-accounts",
            peer_cloud_account,
        )
        if peer_resource_group is not UNDEFINED and peer_resource_group is not None:
            path += self.build_path("peer-resource-groups", peer_resource_group)
        path += self.build_path("peer-vpcs", peer_vpc)
        if peer_region is not None:
            path += self.build_path("peer-regions", peer_region)
        return self.verify(self.delete, path)

    def get_project_vpc_peering_connection(
        self,
        project,
        project_vpc_id,
        peer_cloud_account,
        peer_vpc,
        peer_region=UNDEFINED,
        peer_resource_group=UNDEFINED,
    ):
        vpc = self.get_project_vpc(project=project, project_vpc_id=project_vpc_id)
        for peering_connection in vpc["peering_connections"]:
            # pylint: disable=too-many-boolean-expressions
            if (
                peering_connection["peer_cloud_account"] == peer_cloud_account and peering_connection["peer_vpc"] == peer_vpc
                and (peer_region is UNDEFINED or peering_connection["peer_region"] == peer_region)
                and (peer_resource_group is UNDEFINED or peering_connection["peer_resource_group"] == peer_resource_group)
            ):
                return peering_connection
        if peer_resource_group is not UNDEFINED and peer_resource_group is not None:
            peer_resource_group_msg = " in resource group {}".format(peer_resource_group)
        else:
            peer_resource_group_msg = ""
        if peer_region is not UNDEFINED and peer_region is not None:
            peer_region_msg = " in region {}".format(peer_region)
        else:
            peer_region_msg = ""
        msg = "Peering connection to peer account {}{} VPC {}{} does not exist".format(
            peer_cloud_account,
            peer_resource_group_msg,
            peer_vpc,
            peer_region_msg,
        )
        raise KeyError(msg)

    def update_project_vpc_user_peer_network_cidrs(self, project, project_vpc_id, add=None, delete=None):
        path = self.build_path(
            "project",
            project,
            "vpcs",
            project_vpc_id,
            "user-peer-network-cidrs",
        )
        body = {}
        if add:
            body["add"] = add
        if delete:
            body["delete"] = delete
        return self.verify(self.put, path, body=body)

    def list_project_tags(self, project):
        path = self.build_path(
            "project",
            project,
            "tags",
        )
        return self.verify(self.get, path)

    def update_project_tags(self, project, tag_updates):
        path = self.build_path(
            "project",
            project,
            "tags",
        )
        body = {"tags": tag_updates}
        return self.verify(self.patch, path, body=body)

    def replace_project_tags(self, project, tags):
        path = self.build_path(
            "project",
            project,
            "tags",
        )
        body = {"tags": tags}
        return self.verify(self.put, path, body=body)

    def create_service(
        self,
        project,
        service,
        service_type,
        plan,
        disk_space_mb=None,
        cloud=None,
        user_config=None,
        project_vpc_id=UNDEFINED,
        service_integrations=None,
        termination_protection=False,
        static_ips=(),
    ):
        user_config = user_config or {}
        body = {
            "cloud": cloud,
            "plan": plan,
            "service_integrations": service_integrations,
            "service_name": service,
            "service_type": service_type,
            "user_config": user_config,
            "termination_protection": termination_protection,
            "static_ips": static_ips,
        }
        if disk_space_mb is not None:
            body["disk_space_mb"] = disk_space_mb
        if project_vpc_id is not UNDEFINED:
            body["project_vpc_id"] = project_vpc_id
        return self.verify(
            self.post,
            self.build_path("project", project, "service"),
            body=body,
            result_key="service",
        )

    def update_service(
        self,
        project,
        service,
        cloud=None,
        maintenance=None,
        user_config=None,
        plan=None,
        disk_space_mb=None,
        powered=None,
        termination_protection=None,
        project_vpc_id=UNDEFINED,
    ):
        user_config = user_config or {}
        body = {}
        if cloud is not None:
            body["cloud"] = cloud
        if maintenance is not None:
            body["maintenance"] = maintenance
        if plan is not None:
            body["plan"] = plan
        if disk_space_mb is not None:
            body["disk_space_mb"] = disk_space_mb
        if powered is not None:
            body["powered"] = powered
        if user_config is not None:
            body["user_config"] = user_config
        if project_vpc_id is not UNDEFINED:
            body["project_vpc_id"] = project_vpc_id
        if termination_protection is not None:
            body["termination_protection"] = termination_protection

        path = self.build_path("project", project, "service", service)
        return self.verify(self.put, path, body=body, result_key="service")

    def reset_service_credentials(self, project, service):
        path = self.build_path("project", project, "service", service, "credentials", "reset")
        return self.verify(self.put, path, result_key="service")

    def _privatelink_path(self, project, service, cloud_provider, *rest):
        return self.build_path("project", project, "service", service, "privatelink", cloud_provider, *rest)

    def create_service_privatelink_aws(self, project, service, principals):
        path = self._privatelink_path(project, service, "aws")
        return self.verify(self.post, path, body={"principals": principals})

    def update_service_privatelink_aws(self, project, service, principals):
        path = self._privatelink_path(project, service, "aws")
        return self.verify(self.put, path, body={"principals": principals})

    def delete_service_privatelink_aws(self, project, service):
        path = self._privatelink_path(project, service, "aws")
        return self.verify(self.delete, path)

    def get_service_privatelink_aws(self, project, service):
        path = self._privatelink_path(project, service, "aws")
        return self.verify(self.get, path)

    def list_service_privatelink_aws_connections(self, project, service):
        path = self._privatelink_path(project, service, "aws") + "/connections"
        return self.verify(self.get, path, result_key="connections")

    def create_service_privatelink_azure(self, project, service, user_subscription_ids):
        path = self._privatelink_path(project, service, "azure")
        return self.verify(self.post, path, body={"user_subscription_ids": user_subscription_ids})

    def refresh_service_privatelink_azure(self, project, service):
        path = self._privatelink_path(project, service, "azure", "refresh")
        return self.verify(self.post, path)

    def update_service_privatelink_azure(self, project, service, user_subscription_ids):
        path = self._privatelink_path(project, service, "azure")
        return self.verify(self.put, path, body={"user_subscription_ids": user_subscription_ids})

    def update_service_privatelink_connection_azure(self, project, service, privatelink_connection_id, user_ip_address):
        path = self._privatelink_path(project, service, "azure", "connections", privatelink_connection_id)
        return self.verify(self.put, path, body={"user_ip_address": user_ip_address})

    def approve_service_privatelink_connection_azure(self, project, service, privatelink_connection_id):
        path = self._privatelink_path(project, service, "azure", "connections", privatelink_connection_id, "approve")
        return self.verify(self.post, path)

    def delete_service_privatelink_azure(self, project, service):
        path = self._privatelink_path(project, service, "azure")
        return self.verify(self.delete, path)

    def get_service_privatelink_azure(self, project, service):
        path = self._privatelink_path(project, service, "azure")
        return self.verify(self.get, path)

    def list_service_privatelink_azure_connections(self, project, service):
        path = self._privatelink_path(project, service, "azure") + "/connections"
        return self.verify(self.get, path, result_key="connections")

    def list_privatelink_cloud_availability(self, project):
        path = self.build_path("project", project, "privatelink-availability")
        return self.verify(self.get, path, result_key="privatelink_availability")

    def _static_ip_address_path(self, project, *parts):
        return self.build_path("project", project, "static-ips", *parts)

    def list_static_ip_addresses(self, project):
        path = self._static_ip_address_path(project)
        return self.verify(self.get, path, result_key="static_ips")

    def create_static_ip_address(self, project, cloud_name):
        path = self._static_ip_address_path(project)
        return self.verify(self.post, path, body={"cloud_name": cloud_name})

    def associate_static_ip_address(self, project, static_ip_id, service_name):
        path = self._static_ip_address_path(project, static_ip_id, "association")
        return self.verify(self.post, path, body={"service_name": service_name})

    def dissociate_static_ip_address(self, project, static_ip_id):
        path = self._static_ip_address_path(project, static_ip_id, "association")
        return self.verify(self.delete, path)

    def delete_static_ip_address(self, project, static_ip_id):
        path = self._static_ip_address_path(project, static_ip_id)
        return self.verify(self.delete, path)

    def delete_service(self, project, service):
        return self.verify(self.delete, self.build_path("project", project, "service", service))

    def get_pg_service_current_queries(self, project, service):
        warnings.warn("Use the get_service_current_queries method", DeprecationWarning)
        return self.get_service_current_queries(project, service)

    def get_pg_service_query_stats(self, project, service):
        warnings.warn("Use the get_service_query_stats method", DeprecationWarning)
        return self.get_service_query_stats(project, service, service_type="pg")

    def reset_pg_service_query_stats(self, project, service):
        warnings.warn("Use the reset_service_query_stats method", DeprecationWarning)
        return self.reset_service_query_stats(project, service)

    def get_service_current_queries(self, project, service):
        path = self.build_path("project", project, "service", service, "query", "activity")
        return self.verify(
            self.post,
            path,
            result_key="queries",
            body={
                "limit": 100,
                "order_by": "query_duration:desc"
            },
        )

    def get_service_query_stats(self, project, service, service_type=None):
        if service_type is None:
            service_type = self.get_service(project, service)["service_type"]
        path = self.build_path("project", project, "service", service, service_type, "query", "stats")
        return self.verify(
            self.post,
            path,
            result_key="queries",
            body={
                "limit": 100,
                "order_by": "calls:desc" if service_type == "pg" else "count_star:desc",
            },
        )

    def reset_service_query_stats(self, project, service):
        path = self.build_path("project", project, "service", service, "query", "stats", "reset")
        return self.verify(self.put, path, result_key="queries")

    def list_service_tags(self, project, service):
        path = self.build_path(
            "project",
            project,
            "service",
            service,
            "tags",
        )
        return self.verify(self.get, path)

    def update_service_tags(self, project, service, tag_updates):
        path = self.build_path(
            "project",
            project,
            "service",
            service,
            "tags",
        )
        body = {"tags": tag_updates}
        return self.verify(self.patch, path, body=body)

    def replace_service_tags(self, project, service, tags):
        path = self.build_path(
            "project",
            project,
            "service",
            service,
            "tags",
        )
        body = {"tags": tags}
        return self.verify(self.put, path, body=body)

    def get_services(self, project):
        return self.verify(
            self.get,
            self.build_path("project", project, "service"),
            result_key="services",
        )

    def get_service_types(self, project):
        if project is None:
            path = "/service_types"
        else:
            path = self.build_path("project", project, "service_types")
        return self.verify(self.get, path, result_key="service_types")

    def create_account(self, account_name):
        body = {
            "account_name": account_name,
        }
        return self.verify(self.post, "/account", body=body, result_key="account")

    def delete_account(self, account_id):
        return self.verify(self.delete, self.build_path("account", account_id))

    def update_account(self, account_id, account_name):
        body = {
            "account_name": account_name,
        }
        return self.verify(
            self.put,
            self.build_path("account", account_id),
            body=body,
            result_key="account",
        )

    def get_accounts(self):
        return self.verify(self.get, "/account", result_key="accounts")

    def create_account_authentication_method(self, account_id, method_name, method_type, options=None):
        body = dict(options) if options else {}
        body["authentication_method_name"] = method_name
        body["authentication_method_type"] = method_type
        path = self.build_path("account", account_id, "authentication")
        return self.verify(self.post, path, body=body, result_key="authentication_method")

    def delete_account_authentication_method(self, account_id, authentication_id):
        return self.verify(
            self.delete,
            self.build_path("account", account_id, "authentication", authentication_id),
        )

    def update_account_authentication_method(
        self,
        account_id,
        authentication_id,
        method_name=None,
        method_enable=None,
        options=None,
    ):
        body = dict(options) if options else {}
        if method_name is not None:
            body["authentication_method_name"] = method_name
        if method_enable is not None:
            body["authentication_method_enabled"] = method_enable

        path = self.build_path("account", account_id, "authentication", authentication_id)
        return self.verify(self.put, path, body=body, result_key="authentication_method")

    def get_account_authentication_methods(self, account_id):
        path = self.build_path("account", account_id, "authentication")
        return self.verify(self.get, path, result_key="authentication_methods")

    def create_project(
        self,
        project,
        account_id=None,
        billing_group_id=None,
        card_id=None,
        cloud=None,
        copy_from_project=None,
        country_code=None,
        billing_address=None,
        billing_currency=None,
        billing_extra_text=None,
        vat_id=None,
        billing_emails=None,
        tech_emails=None,
        use_source_project_billing_group=None,
    ):
        body = {
            "card_id": card_id,
            "cloud": cloud,
            "project": project,
        }
        if account_id is not None:
            body["account_id"] = account_id
        if billing_group_id is not None:
            body["billing_group_id"] = billing_group_id
        if copy_from_project is not None:
            body["copy_from_project"] = copy_from_project
        if country_code is not None:
            body["country_code"] = country_code
        if billing_address is not None:
            body["billing_address"] = billing_address
        if billing_currency is not None:
            body["billing_currency"] = billing_currency
        if billing_extra_text is not None:
            body["billing_extra_text"] = billing_extra_text
        if vat_id is not None:
            body["vat_id"] = vat_id
        if billing_emails is not None:
            body["billing_emails"] = [{"email": email} for email in billing_emails]
        if tech_emails is not None:
            body["tech_emails"] = [{"email": email} for email in tech_emails]
        if use_source_project_billing_group is not None:
            body["use_source_project_billing_group"] = use_source_project_billing_group

        return self.verify(self.post, "/project", body=body, result_key="project")

    def delete_project(self, project):
        return self.verify(self.delete, self.build_path("project", project))

    def get_project(self, project):
        return self.verify(self.get, self.build_path("project", project), result_key="project")

    def get_projects(self):
        return self.verify(self.get, "/project", result_key="projects")

    def update_project(
        self,
        project,
        new_project_name=None,
        account_id=None,
        card_id=None,
        cloud=None,
        country_code=None,
        billing_address=None,
        billing_currency=None,
        billing_extra_text=None,
        vat_id=None,
        billing_emails=None,
        tech_emails=None,
    ):
        body = {}
        if new_project_name is not None:
            body["project_name"] = new_project_name
        if account_id is not None:
            body["account_id"] = account_id
        if card_id is not None:
            body["card_id"] = card_id
        if cloud is not None:
            body["cloud"] = cloud
        if country_code is not None:
            body["country_code"] = country_code
        if billing_address is not None:
            body["billing_address"] = billing_address
        if billing_currency is not None:
            body["billing_currency"] = billing_currency
        if billing_extra_text is not None:
            body["billing_extra_text"] = billing_extra_text
        if vat_id is not None:
            body["vat_id"] = vat_id
        if billing_emails is not None:
            body["billing_emails"] = [{"email": email} for email in billing_emails]
        if tech_emails is not None:
            body["tech_emails"] = [{"email": email} for email in tech_emails]

        return self.verify(
            self.put,
            self.build_path("project", project),
            body=body,
            result_key="project",
        )

    def get_project_ca(self, project):
        return self.verify(self.get, self.build_path("project", project, "kms", "ca"))

    def get_service_ca(self, project, service, ca):
        path = self.build_path("project", project, "service", service, "kms", "ca", ca)
        return self.verify(self.get, path)

    def get_service_keypair(self, project, service, keypair):
        path = self.build_path("project", project, "service", service, "kms", "keypairs", keypair)
        return self.verify(self.get, path)

    def invite_project_user(self, project, user_email, member_type=None):
        body = {
            "user_email": user_email,
        }
        if member_type is not None:
            body["member_type"] = member_type
        return self.verify(self.post, self.build_path("project", project, "invite"), body=body)

    def remove_project_user(self, project, user_email):
        return self.verify(self.delete, self.build_path("project", project, "user", user_email))

    def list_project_users(self, project):
        return self.verify(self.get, self.build_path("project", project, "users"), result_key="users")

    def list_invited_project_users(self, project):
        return self.verify(self.get, self.build_path("project", project, "users"), result_key="invitations")

    def create_user(self, email, password, real_name, *, tenant=None):
        request = {
            "email": email,
            "real_name": real_name,
        }
        if tenant is not None:
            request["tenant"] = tenant
        if password is not None:
            request["password"] = password
        return self.verify(self.post, "/user", body=request)

    def get_user_info(self):
        return self.verify(self.get, "/me", result_key="user")

    def access_token_create(self, description, extend_when_used=False, max_age_seconds=None):
        request = {
            "description": description,
            "extend_when_used": extend_when_used,
            "max_age_seconds": max_age_seconds,
        }
        return self.verify(self.post, "/access_token", body=request)

    def access_token_revoke(self, token_prefix):
        return self.verify(self.delete, self.build_path("access_token", token_prefix))

    def access_token_update(self, token_prefix, description):
        request = {"description": description}
        return self.verify(self.put, self.build_path("access_token", token_prefix), body=request)

    def access_tokens_list(self):
        return self.verify(self.get, "/access_token", result_key="tokens")

    def expire_user_tokens(self):
        return self.verify(self.post, "/me/expire_tokens")

    def get_service_logs(self, project, service, sort_order=None, offset=None, limit=100):
        body = {"limit": limit}
        if offset is not None:
            body["offset"] = str(offset)
        if sort_order is not None:
            body["sort_order"] = sort_order
        return self.verify(
            self.post,
            self.build_path("project", project, "service", service, "logs"),
            body=body,
        )

    def get_events(self, project, limit=100):
        params = {"limit": limit}
        return self.verify(
            self.get,
            self.build_path("project", project, "events"),
            params=params,
            result_key="events",
        )

    def get_cards(self):
        return self.verify(self.get, "/card", result_key="cards")

    def add_card(self, stripe_token):
        request = {
            "stripe_token": stripe_token,
        }
        return self.verify(self.post, "/card", body=request, result_key="card")

    def update_card(self, card_id, **kwargs):
        keys = {"exp_month", "exp_year", "name"}
        wrong = set(kwargs) - keys
        assert not wrong, "invalid arguments to update_card: {!r}".format(wrong)
        request = {}
        for key in keys:
            value = kwargs.get(key)
            if value is not None:
                if key in {"exp_month", "exp_year"}:
                    expected = int
                else:
                    expected = str

                assert isinstance(value, expected), "expected '{}' type for argument '{}'".format(expected, key)

                request[key] = value

        return self.verify(self.put, self.build_path("card", card_id), body=request, result_key="card")

    def remove_card(self, card_id):
        return self.verify(self.delete, self.build_path("card", card_id))

    def get_stripe_key(self):
        return self.verify(self.get, self.build_path("config", "stripe_key"))

    def list_project_credits(self, project):
        return self.verify(
            self.get,
            self.build_path("project", project, "credits"),
            result_key="credits",
        )

    def claim_project_credit(self, project, credit_code):
        return self.verify(
            self.post,
            self.build_path("project", project, "credits"),
            body={"code": credit_code},
            result_key="credit",
        )

    def create_billing_group(
        self,
        billing_group_name,
        *,
        account_id=None,
        card_id=None,
        vat_id=None,
        billing_currency=None,
        billing_extra_text=None,
        billing_emails=None,
        company=None,
        address_lines=None,
        country_code=None,
        city=None,
        state=None,
        zip_code=None,
    ):
        body = {"billing_group_name": billing_group_name}
        if account_id is not None:
            body["account_id"] = account_id
        if card_id is not None:
            body["card_id"] = card_id
        if vat_id is not None:
            body["vat_id"] = vat_id
        if billing_currency is not None:
            body["billing_currency"] = billing_currency
        if billing_extra_text is not None:
            body["billing_extra_text"] = billing_extra_text
        if billing_emails is not None:
            body["billing_emails"] = [{"email": email} for email in billing_emails]
        if company is not None:
            body["company"] = company
        if address_lines is not None:
            body["address_lines"] = address_lines
        if country_code is not None:
            body["country_code"] = country_code
        if city is not None:
            body["city"] = city
        if state is not None:
            body["state"] = state
        if zip_code is not None:
            body["zip_code"] = zip_code

        return self.verify(self.post, "/billing-group", body=body, result_key="billing_group")

    def get_billing_groups(self):
        return self.verify(self.get, "/billing-group", result_key="billing_groups")

    def get_billing_group(self, billing_group):
        return self.verify(self.get, self.build_path("billing-group", billing_group), result_key="billing_group")

    def get_billing_group_projects(self, billing_group):
        return self.verify(self.get, self.build_path("billing-group", billing_group, "projects"), result_key="projects")

    def update_billing_group(
        self,
        billing_group,
        *,
        billing_group_name=None,
        account_id=None,
        card_id=None,
        vat_id=None,
        billing_currency=None,
        billing_extra_text=None,
        billing_emails=None,
        company=None,
        address_lines=None,
        country_code=None,
        city=None,
        state=None,
        zip_code=None,
    ):
        body = {}
        if billing_group_name is not None:
            body["billing_group_name"] = billing_group_name
        if account_id is not None:
            body["account_id"] = account_id
        if card_id is not None:
            body["card_id"] = card_id
        if vat_id is not None:
            body["vat_id"] = vat_id
        if billing_currency is not None:
            body["billing_currency"] = billing_currency
        if billing_extra_text is not None:
            body["billing_extra_text"] = billing_extra_text
        if billing_emails is not None:
            body["billing_emails"] = [{"email": email} for email in billing_emails]
        if company is not None:
            body["company"] = company
        if address_lines is not None:
            body["address_lines"] = address_lines
        if country_code is not None:
            body["country_code"] = country_code
        if city is not None:
            body["city"] = city
        if state is not None:
            body["state"] = state
        if zip_code is not None:
            body["zip_code"] = zip_code

        return self.verify(self.put, self.build_path("billing-group", billing_group), body=body, result_key="billing_group")

    def delete_billing_group(self, billing_group):
        return self.verify(self.delete, self.build_path("billing-group", billing_group))

    def assign_projects_to_billing_group(self, billing_group, *, project_names):
        body = {"projects_names": project_names}
        self.verify(self.post, self.build_path("billing-group", billing_group, "projects-assign"), body=body)

    def get_billing_group_events(self, billing_group, *, limit=100):
        return self.verify(
            self.get,
            self.build_path("billing-group", billing_group, "events"),
            params={"limit": limit},
            result_key="events"
        )

    def list_billing_group_credits(self, billing_group):
        return self.verify(
            self.get,
            self.build_path("billing-group", billing_group, "credits"),
            result_key="credits",
        )

    def claim_billing_group_credit(self, billing_group, *, credit_code):
        return self.verify(
            self.post,
            self.build_path("billing-group", billing_group, "credits"),
            body={"code": credit_code},
            result_key="credit",
        )

    def list_billing_group_invoices(self, billing_group, *, sort=None):
        params = {"sort": sort} if sort else {}
        invoices = self.verify(
            self.get, self.build_path("billing-group", billing_group, "invoice"), params=params, result_key="invoices"
        )
        for invoice in invoices:
            if invoice.get("download_cookie"):
                invoice["download"] = self.base_url + self.api_prefix + self.build_path(
                    "billing-group", billing_group, "invoice", invoice["invoice_number"], invoice["download_cookie"]
                )
        return invoices

    def get_billing_group_invoice_lines(self, billing_group, invoice_number):
        return self.verify(
            self.get,
            self.build_path("billing-group", billing_group, "invoice", invoice_number, "lines"),
            result_key="lines"
        )

    def start_service_maintenance(self, project, service):
        return self.verify(
            self.put,
            self.build_path("project", project, "service", service, "maintenance", "start"),
        )

    def create_ticket(self, project, severity, title, description, service=None):
        body = {
            "severity": severity,
            "title": title,
            "description": description,
        }
        if service:
            body["service_name"] = service

        return self.verify(self.post, self.build_path("project", project, "tickets"), body=body)

    def list_tickets(self, project):
        return self.verify(self.get, self.build_path("project", project, "tickets"))

    def get_service_migration_status(self, project, service):
        return self.verify(
            self.get,
            self.build_path("project", project, "service", service, "migration"),
        )

    def list_teams(self, account_id):
        return self.verify(
            self.get,
            self.build_path("account", account_id, "teams"),
            result_key="teams",
        )

    def create_team(self, account_id, team_name):
        return self.verify(
            self.post,
            self.build_path("account", account_id, "teams"),
            body={"team_name": team_name},
        )

    def delete_team(self, account_id, team_id):
        return self.verify(self.delete, self.build_path("account", account_id, "team", team_id))

    def list_team_members(self, account_id, team_id):
        return self.verify(
            self.get,
            self.build_path("account", account_id, "team", team_id, "members"),
            result_key="members",
        )

    def add_team_member(self, account_id, team_id, email):
        return self.verify(
            self.post,
            self.build_path("account", account_id, "team", team_id, "members"),
            body={"email": email},
        )

    def list_team_invites(self, account_id, team_id):
        return self.verify(
            self.get,
            self.build_path("account", account_id, "team", team_id, "invites"),
            result_key="account_invites",
        )

    def delete_team_member(self, account_id, team_id, user_id):
        return self.verify(
            self.delete,
            self.build_path("account", account_id, "team", team_id, "member", user_id),
        )

    def list_team_projects(self, account_id, team_id):
        return self.verify(
            self.get,
            self.build_path("account", account_id, "team", team_id, "projects"),
            result_key="projects",
        )

    def attach_team_to_project(self, account_id, team_id, project, team_type):
        return self.verify(
            self.post,
            self.build_path("account", account_id, "team", team_id, "project", project),
            body={"team_type": team_type},
        )

    def detach_team_from_project(self, account_id, team_id, project):
        return self.verify(
            self.delete,
            self.build_path("account", account_id, "team", team_id, "project", project),
        )



# Copyright 2019, Aiven, https://aiven.io/
#
# This file is under the Apache License, Version 2.0.
# See the file `LICENSE` for details.
from aiven.client.pretty import format_item

import datetime
import decimal
import pytest

pytestmark = [pytest.mark.unittest, pytest.mark.all]


@pytest.mark.parametrize(
    "value,expected",
    [
        (1, "1"),
        ("a_string", "a_string"),
        (datetime.datetime(year=2019, month=12, day=23), "2019-12-23T00:00:00"),
        ([datetime.datetime(year=2019, month=12, day=23)], "2019-12-23T00:00:00"),
        (
            ["x", datetime.datetime(year=2019, month=12, day=23)],
            "x, 2019-12-23T00:00:00",
        ),
        (decimal.Decimal("64.23"), "64.23"),
        (
            {
                "a": decimal.Decimal("12.34"),
                "b": datetime.datetime(year=2019, month=12, day=23),
            },
            '{"a": "12.34", "b": "2019-12-23T00:00:00"}',
        ),
    ],
)
def test_format_item(value, expected):
    assert format_item(None, value) == expected



# Copyright (c) 2021 Aiven, Helsinki, Finland. https://aiven.io/
from ._utils import find_component, find_user
from .common import ConnectionInfoError, Store


class KafkaConnectionInfo:  # pylint: disable=too-few-public-methods
    def __init__(
        self,
        host,
        port,
    ):
        self.host = host
        self.port = port

    def _kafkacat(self, protocol, ca_path, extra, store: Store, get_project_ca):
        store.handle(get_project_ca, ca_path)
        address = f"{self.host}:{self.port}"
        return ["kafkacat", "-b", address, "-X", f"security.protocol={protocol}", "-X", f"ssl.ca.location={ca_path}", *extra]


class KafkaCertificateConnectionInfo(KafkaConnectionInfo):
    def __init__(self, host, port, client_cert, client_key):
        super().__init__(host, port)
        self.client_cert = client_cert
        self.client_key = client_key

    def kafkacat(self, store, get_project_ca, ca_path, client_key_path, client_cert_path):
        store.handle(lambda: self.client_cert, client_cert_path)
        store.handle(lambda: self.client_key, client_key_path)

        extra = [
            "-X",
            f"ssl.key.location={client_key_path}",
            "-X",
            f"ssl.certificate.location={client_cert_path}",
        ]
        return self._kafkacat("SSL", ca_path, extra, store, get_project_ca)

    @classmethod
    def from_service(
        cls,
        service,
        *,
        route,
        privatelink_connection_id,
        username,
    ):
        if service["service_type"] != "kafka":
            raise ConnectionInfoError(
                "Cannot format kafka connection info for service type {service_type}".format_map(service)
            )

        info = find_component(
            service["components"],
            route=route,
            privatelink_connection_id=privatelink_connection_id,
            kafka_authentication_method="certificate"
        )
        user = find_user(service, username)
        if "access_cert" not in user:
            raise ConnectionInfoError(f"Could not find client certificate for username {username}")
        if "access_key" not in user:
            raise ConnectionInfoError(f"Could not find client key for username {username}")

        client_cert = user["access_cert"]
        client_key = user["access_key"]
        return cls(host=info["host"], port=info["port"], client_cert=client_cert, client_key=client_key)


class KafkaSASLConnectionInfo(KafkaConnectionInfo):
    def __init__(self, host, port, username, password):
        super().__init__(host, port)
        self.username = username
        self.password = password

    @classmethod
    def from_service(
        cls,
        service,
        *,
        route,
        privatelink_connection_id,
        username,
    ):
        if service["service_type"] != "kafka":
            raise ConnectionInfoError(
                "Cannot format kafka connection info for service type {service_type}".format_map(service)
            )

        info = find_component(
            service["components"],
            route=route,
            privatelink_connection_id=privatelink_connection_id,
            kafka_authentication_method="sasl"
        )
        user = find_user(service, username)
        if "password" not in user:
            raise ConnectionInfoError(f"Could not find password for username {username}")
        return cls(host=info["host"], port=info["port"], username=username, password=user["password"])

    def kafkacat(self, store, get_project_ca, ca_path):
        extra = [
            "-X",
            "sasl.mechanisms=SCRAM-SHA-256",
            "-X",
            f"sasl.username={self.username}",
            "-X",
            f"sasl.password={self.password}",
        ]
        return self._kafkacat("SASL_SSL", ca_path, extra, store, get_project_ca)



# Copyright 2015, Aiven, https://aiven.io/
#
# This file is under the Apache License, Version 2.0.
# See the file `LICENSE` for details.

# pylint: disable=no-member
from aiven.client.argx import CommandLineTool
from aiven.client.cliarg import arg


def test_user_config_json_error_json():
    """Test that @arg.user_config_json causes
    CommandLineTool.run() to exit cleanly with return value 1
    if JSON is incorrect
    """

    class T(CommandLineTool):
        """Test class"""

        @arg.user_config_json()
        @arg()
        def t(self):
            """t"""

    error_json_arg = ["t", "--user-config-json", "foo"]
    test_class = T("avn")
    ret = test_class.run(args=error_json_arg)
    assert ret == 1


def test_user_config_json_error_conflict():
    """Test that @arg.user_config_json causes
    CommandLineTool.run() to exit cleanly with return value 1
    if both user_config (-c) and --user-config-json parameters
    are given
    """

    class T(CommandLineTool):
        """Test class"""

        @arg.user_config
        @arg.user_config_json()
        @arg()
        def t(self):
            """t"""

    error_conflict_arg = ["t", "-c", "userconfkey=val", "--user-config-json", '{"foo":"bar"}']
    test_class = T("avn")
    ret = test_class.run(args=error_conflict_arg)
    assert ret == 1


def test_user_config_json_success():
    """Success scenario
    """

    class T(CommandLineTool):
        """Test class"""

        @arg.user_config_json()
        @arg()
        def t(self):
            """t"""

    valid_json_arg = ["t", "--user-config-json", '{"foo":"bar"}']
    test_class = T("avn")
    ret = test_class.run(args=valid_json_arg)
    assert ret is None  # Should run() return 0 actually?
    assert test_class.args.user_config_json == {"foo": "bar"}


def test_user_config_success():
    """Test that user config parameter -c works and not cause conflict with
    --user_config_json
    """

    class T(CommandLineTool):
        """Test class"""

        @arg.user_config
        @arg.user_config_json()
        @arg()
        def t(self):
            """t"""

    user_config_arg = ["t", "-c", "userconfkey=val"]
    test_class = T("avn")
    ret = test_class.run(args=user_config_arg)
    assert ret is None



# Copyright 2015, Aiven, https://aiven.io/
#
# This file is under the Apache License, Version 2.0.
# See the file `LICENSE` for details.

# pylint: disable=no-member
from aiven.client import AivenClient
from aiven.client.cli import AivenCLI, ClientFactory, EOL_ADVANCE_WARNING_TIME
from collections import namedtuple
from datetime import datetime, timedelta, timezone
from unittest import mock

import pytest

pytestmark = [pytest.mark.unittest, pytest.mark.all]


def test_cli():
    with pytest.raises(SystemExit) as excinfo:
        AivenCLI().run(args=["--help"])
    assert excinfo.value.code == 0


def test_cloud_list():
    AivenCLI().run(args=["cloud", "list"])


def test_service_plans():
    AivenCLI().run(args=["service", "plans"])


def test_service_types_v():
    AivenCLI().run(args=["service", "types", "-v"])


def test_service_user_create():
    AivenCLI().run(args=["service", "user-create", "service", "--username", "username"])


def test_service_topic_create():
    AivenCLI().run(args=["service", "topic-create", "--partitions", "42", "--replication", "42", "service1", "topic1"])


def test_service_topic_create_with_tags():
    AivenCLI().run(
        args=[
            "service", "topic-create", "--partitions", "42", "--replication", "42", "--tag", "key-_1=value1", "--tag",
            "key2=az,.0-9_", "service1", "topic1"
        ]
    )


def test_service_topic_update():
    AivenCLI().run(
        args=[
            "service", "topic-update", "--partitions", "42", "--untag", "key-_1", "--untag", "key123", "--tag",
            "key3=az,.0-9_", "--tag", "key234=foo", "service1", "topic1"
        ]
    )


def test_help():
    AivenCLI().run(args=["help"])


def test_create_user_config():
    cli = AivenCLI()
    cli.args = namedtuple("args", ["user_config", "user_option_remove"])
    cli.args.user_config = ["first.second.third=1", "first.second.with.dot=2", "main=3"]
    cli.args.user_option_remove = ["first.second.thirdaway", "foo"]
    schema = {
        "type": "object",
        "properties": {
            "first": {
                "type": "object",
                "properties": {
                    "second": {
                        "type": "object",
                        "properties": {
                            "third": {
                                "type": "integer",
                            },
                            "thirdaway": {
                                "type": ["null", "integer"],
                            },
                            "with.dot": {
                                "type": "integer",
                            },
                        },
                    }
                },
            },
            "foo": {
                "type": ["integer", "null"],
            },
            "main": {
                "type": "integer",
            },
        },
    }
    config = cli.create_user_config(schema)
    assert config == {
        "first": {
            "second": {
                "third": 1,
                "thirdaway": None,
                "with.dot": 2,
            }
        },
        "foo": None,
        "main": 3,
    }


def patched_get_auth_token() -> str:
    return "token"


def build_aiven_cli(client: AivenClient) -> AivenCLI:
    cli = AivenCLI(client_factory=mock.Mock(spec_set=ClientFactory, return_value=client))
    cli._get_auth_token = patched_get_auth_token  # pylint: disable=protected-access
    return cli


def test_service_task_create_migration_check():
    aiven_client = mock.Mock(spec_set=AivenClient)
    aiven_client.create_service_task.return_value = {
        "message": "created",
        "task": {
            "create_time": "2021-08-13T08:00:45Z",
            "result": "",
            "success": None,
            "task_id": "79803598-d09a-44bf-ae2b-34aad942f4e8",
            "task_type": "mysql_migration_check"
        }
    }

    build_aiven_cli(aiven_client).run(
        args=[
            "service", "task-create", "--operation", "migration_check", "--project", "test", "--source-service-uri",
            "mysql://root:password@source-mysql-server:3306/", "--ignore-dbs", "db1", "target-mysql-service-1"
        ]
    )
    aiven_client.create_service_task.assert_called_with(
        project='test',
        service='target-mysql-service-1',
        body={
            'task_type': 'migration_check',
            'migration_check': {
                'source_service_uri': 'mysql://root:password@source-mysql-server:3306/',
                'ignore_dbs': 'db1'
            }
        }
    )


def test_service_task_get_migration_check():
    aiven_client = mock.Mock(spec_set=AivenClient)
    aiven_client.get_service_task.return_value = {
        "create_time": "2021-08-13T07:10:35Z",
        "ignore_dbs": "defaultdb",
        "result": "aiven_mysql_migrate.exceptions.NothingToMigrateException: No databases to migrate",
        "success": False,
        "task_id": "f25eac03-25f7-4de2-be6a-eb476fec1730",
        "task_type": "mysql_migration_check"
    }

    build_aiven_cli(aiven_client).run(
        args=[
            "service", "task-get", "--task-id", "f25eac03-25f7-4de2-be6a-eb476fec1730", "--project", "test",
            "target-mysql-service-1"
        ]
    )
    aiven_client.get_service_task.assert_called_with(
        project='test', service='target-mysql-service-1', task_id='f25eac03-25f7-4de2-be6a-eb476fec1730'
    )


def test_version_eol_check():
    fake_eol_time = datetime(2021, 9, 10, tzinfo=timezone.utc)

    fake_time_safe = fake_eol_time - EOL_ADVANCE_WARNING_TIME - timedelta(days=1)
    fake_time_soon = fake_eol_time - EOL_ADVANCE_WARNING_TIME + timedelta(days=1)

    aiven_client = mock.Mock(spec_set=AivenClient)
    service_type = "pg"
    service_version = "9.6"
    aiven_client.get_service_versions.return_value = [{
        "service_type": service_type,
        "major_version": service_version,
        "aiven_end_of_life_time": "2021-09-10T00:00:00Z",
        "availability_end_time": "2021-06-12T00:00:00Z"
    }]

    cli = AivenCLI(aiven_client)
    cli.client = aiven_client
    cli.confirm = mock.Mock()

    # Test current time < EOL_WARNING time
    with mock.patch("aiven.client.cli.get_current_date", return_value=fake_time_safe):
        cli._do_version_eol_check(service_type, service_version)  # pylint: disable=protected-access
        cli.confirm.assert_not_called()  # No confirmation should have been asked

    # Test current time > EOL_WARNING
    with mock.patch("aiven.client.cli.get_current_date", return_value=fake_time_soon):
        cli._do_version_eol_check(service_type, service_version)  # pylint: disable=protected-access
        cli.confirm.assert_called()  # Confirmation should have been asked



# Copyright 2015, Aiven, https://aiven.io/
#
# This file is under the Apache License, Version 2.0.
# See the file `LICENSE` for details.

from .argx import arg, UserError
from functools import wraps

import json as jsonlib
import os


def get_json_config(path_or_string):
    # If parameter is empty, return an empty dict
    if not path_or_string:
        return {}
    if path_or_string.startswith("@"):
        filepath = path_or_string[1:]
        with open(filepath, "r") as config_file:
            return jsonlib.load(config_file)

    return jsonlib.loads(path_or_string)


def json_path_or_string(param_name):
    def wrapper(fun):
        arg(
            param_name,
            help="JSON string or path (preceded by '@') to a JSON configuration file",
        )(fun)

        @wraps(fun)
        def wrapped(self):
            setattr(
                self.args,
                param_name,
                get_json_config(getattr(self.args, param_name, "")),
            )
            return fun(self)

        return wrapped

    return wrapper


def user_config_json():
    """User config that accepts arbitrary JSON"""

    def wrapper(fun):
        arg(
            "--user-config-json",
            default=None,
            dest="user_config_json",
            help="JSON string or path (preceded by '@') to a JSON configuration file",
        )(fun)

        @wraps(fun)
        def wrapped(self):
            if ("user_config" in self.args and (self.args.user_config_json and self.args.user_config)):
                raise UserError("-c (user config) and --user-config-json parameters can not be used at the same time")
            try:
                setattr(
                    self.args,
                    "user_config_json",
                    get_json_config(self.args.user_config_json),
                )
            except jsonlib.decoder.JSONDecodeError as err:
                raise UserError from err
            return fun(self)

        return wrapped

    return wrapper


arg.account_id = arg("account_id", help="Account identifier")
arg.authentication_id = arg("authentication_id", help="Account authentication method identifier")
arg.billing_address = arg("--billing-address", help="Physical billing address for invoices")
arg.billing_currency = arg("--billing-currency", help="Currency for charges")
arg.billing_extra_text = arg(
    "--billing-extra-text",
    help="Extra text to include in invoices (e.g. cost center id)",
)
arg.billing_group = arg("id", help="Billing group ID")
arg.card_id = arg("--card-id", help="Card ID")
arg.cloud = arg("--cloud", help="Cloud to use (see 'cloud list' command)")
arg.cloud_mandatory = arg("--cloud", help="Cloud to use (see 'cloud list' command)", required=True)
arg.config_cmdline = arg(
    "-c",
    dest="config_cmdline",
    metavar="KEY=VALUE",
    action="append",
    default=[],
    help="Additional configuration option in the form name=value",
)
arg.config_file = arg(
    "-f",
    dest="config_file",
    metavar="KEY=VALUE",
    action="append",
    default=[],
    help="Additional configuration option whose value is loaded from file in the form name=filename",
)
arg.country_code = arg("--country-code", help="Billing country code")
arg.disk_space_mb = arg(
    "--disk-space-gib", dest="disk_space_mb", type=lambda value: int(value) * 1024, help="Disk space for data storage (GiB)"
)
arg.email = arg("email", help="User email address")
arg.force = arg(
    "-f",
    "--force",
    help="Force action without interactive confirmation",
    action="store_true",
    default=False,
)
arg.index_name = arg("index_name", help="Index name")
arg.json = arg("--json", help="Raw json output", action="store_true", default=False)
arg.min_insync_replicas = arg(
    "--min-insync-replicas",
    type=int,
    help="Minimum required nodes In Sync Replicas (ISR) to produce to a partition (default: 1)",
)
arg.partitions = arg("--partitions", type=int, required=True, help="Number of partitions")
arg.project = arg(
    "--project",
    help="Project name to use, default %(default)r",
    default=os.environ.get("AIVEN_PROJECT"),
)
arg.replication = arg("--replication", type=int, required=True, help="Replication factor")
arg.retention = arg("--retention", type=int, help="Retention period in hours (default: unlimited)")
arg.retention_bytes = arg("--retention-bytes", type=int, help="Retention limit in bytes (default: unlimited)")
arg.tag = arg(
    "--tag", dest="topic_option_tag", metavar="KEY[=VALUE]", action="append", help="Tag to add into topic metadata"
)
arg.tagupdate = arg(
    "--tag",
    dest="topic_option_tag",
    metavar="KEY[=VALUE]",
    action="append",
    help="Tag to add or replace into topic metadata"
)
arg.untag = arg(
    "--untag", dest="topic_option_untag", metavar="KEY", action="append", help="Tag to delete from topic metadata"
)
arg.service_name = arg("service_name", help="Service name")
arg.service_name_mandatory = arg("service_name", help="Service name", required=True)
arg.service_type = arg("-t", "--service-type", help="Type of service (see 'service list-types')")
arg.static_ip_id = arg("static_ip_id", help="Static IP address ID")
arg.ns_name = arg("ns_name", help="Namespace name")
arg.ns_type = arg("--ns-type", help="Namespace type ('aggregated' or 'unaggregated')", required=True)
arg.ns_retention_mandatory = arg(
    "--ns-retention", help="Namespace retention period (written like 30m/25h etc)", required=True
)
arg.ns_retention = arg("--ns-retention", help="Namespace retention period (written like 30m/25h etc)", required=False)
arg.ns_resolution = arg("--ns-resolution", help="Namespace resolution (written like 30m/25h etc)")
arg.ns_blocksize_dur = arg("--ns-blocksize-dur", help="Namespace block size duration (written like 30m/25h etc)")
arg.ns_block_data_expiry_dur = arg(
    "--ns-block-data-expiry-dur", help="Namespace block size duration (written like 30m/25h etc)"
)
arg.ns_buffer_future_dur = arg("--ns-buffer-future-dur", help="Namespace block size duration (written like 30m/25h etc)")
arg.ns_buffer_past_dur = arg("--ns-buffer-past-dur", help="Namespace block size duration (written like 30m/25h etc)")
arg.ns_writes_to_commitlog = arg("--ns-writes-to-commitlog", help="Namespace writes to commit log")
arg.team_name = arg("--team-name", help="Team  name", required=True)
arg.team_id = arg("--team-id", help="Team identifier", required=True)
arg.timeout = arg("--timeout", type=int, help="Wait for up to N seconds (default: infinite)")
arg.topic = arg("topic", help="Topic name")
arg.user_config = arg(
    "-c",
    dest="user_config",
    metavar="KEY=VALUE",
    action="append",
    default=[],
    help="Apply a configuration setting. See 'avn service types -v' for available values.",
)
arg.user_config_json = user_config_json
arg.user_id = arg("--user-id", help="User identifier", required=True)
arg.user_option_remove = arg(
    "--remove-option",
    dest="user_option_remove",
    action="append",
    default=[],
    help="Remove a configuration setting. See 'avn service types -v' for available settings.",
)
arg.vat_id = arg("--vat-id", help="VAT ID of an EU VAT area business")
arg.verbose = arg("-v", "--verbose", help="Verbose output", action="store_true", default=False)
arg.connector_name = arg("connector", help="Connector name")
arg.json_path_or_string = json_path_or_string
arg.subject = arg("--subject", required=True, help="Subject name")
arg.version_id = arg("--version-id", required=True, help="Subject version")
arg.compatibility = arg(
    "--compatibility",
    required=True,
    choices=["BACKWARD", "FORWARD", "FULL", "NONE"],
    help="Choose a compatibility level for the subject",
)
arg.schema = arg("--schema", required=True, help="Schema string quote escaped")

arg.source_cluster = arg("-s", "--source-cluster", required=True, help="Source cluster alias")
arg.target_cluster = arg("-t", "--target-cluster", required=True, help="Target cluster alias")

arg.billing_email = arg("--billing-email", action="append", help="Billing email address")
arg.tech_email = arg("--tech-email", action="append", help="Tech email address")
